
' Written by Colin MacDonald
' Copyright (C) 2002-2009 Nikolaus Gebhardt
' This file is part of the "Irrlicht Engine".
' For conditions of distribution And use, see copyright notice in irrlicht.h


SuperStrict
Framework Irrlicht.Core

Rem
    Normally, you are limited To 8 dynamic lights per scene: this is a hardware limit.  If you
    want To use more dynamic lights in your scene, Then you can register an optional light
    manager that allows you To To turn lights on And off at specific point during rendering.
    You are still limited To 8 lights, but the limit is per scene node.

    This is completely optional: If you do Not register a light manager, Then a Default
    distance-based scheme will be used To prioritise hardware lights based on their distance
    from the active camera.

	NO_MANAGEMENT disables the light manager And shows Irrlicht's default light behaviour.
    The 8 lights nearest To the camera will be turned on, And other lights will be turned off.
    In this example, this produces a funky looking but incoherent light display.

	LIGHTS_NEAREST_NODE shows an implementation that turns on a limited number of lights
    per mesh scene node.  If finds the 3 lights that are nearest To the node being rendered,
    And turns them on, turning all other lights off.  This works, but as it operates on every
    light For every node, it does Not scale well with many lights.  The flickering you can see
    in this demo is due To the lights swapping their relative positions from the cubes
    (a deliberate demonstration of the limitations of this technique).

	LIGHTS_IN_ZONE shows a technique For turning on lights based on a 'zone'. Each empty scene
    node is considered To be the parent of a zone.  When nodes are rendered, they turn off all
    lights, Then find their parent 'zone' and turn on all lights that are inside that zone, i.e.
	are  descendents of it in the scene graph.  This produces True 'local' lighting for each cube
    in this example.  You could use a similar technique To locally light all meshes in (e.g.)
    a room, without the lights spilling out To other rooms.

	This light manager is also an event receiver this is purely For simplicity in this example,
    it's neither necessary nor recommended for a real application.
EndRem

Type CMyLightManager Extends ILightManager

	' LightManagementMode enum
	Const NO_MANAGEMENT:Int = 0
	Const LIGHTS_NEAREST_NODE:Int = 1
	Const LIGHTS_IN_ZONE:Int = 2

	Field Mode:Int = NO_MANAGEMENT
	Field RequestedMode:Int = NO_MANAGEMENT

	' These data represent the state information that this light manager
	' is interested in.
	Field SceneManager:ISceneManager
	Field SceneLightList:ILightSceneNode[]
	Field CurrentRenderPass:Int = ESNRP_NONE ' E_SCENE_NODE_RENDER_PASS
	Field CurrentSceneNode:ISceneNode

	' This is called before the first scene node is rendered.
	Method OnPreRender(lightList:ILightSceneNode[])
		' Update the mode changing it here ensures that it's consistent throughout a render
		Mode = RequestedMode

		' Store the light list. I am free To alter this list Until the End of OnPostRender().
		SceneLightList = lightList
	EndMethod

	' Called after the last scene node is rendered.
	Method OnPostRender()
		' Since light management might be switched off in the event handler, we'll turn all
		' lights on To ensure that they are in a consistent state. You wouldn't normally have
		' To do this when using a light manager, since you'd continue to do light management
		' yourself.
		For Local light:ILightSceneNode = EachIn SceneLightList
			light.setVisible(True)
		Next
	EndMethod

	Method OnRenderPassPreRender(E_SCENE_NODE_RENDER_PASS:Int) 
		' I don't have to do anything here except remember which render pass I am in.
		CurrentRenderPass = E_SCENE_NODE_RENDER_PASS
	EndMethod
	
	Method OnRenderPassPostRender(E_SCENE_NODE_RENDER_PASS:Int) 
		' I only want solid nodes To be lit, so after the solid pass, turn all lights off.
		If (ESNRP_SOLID = E_SCENE_NODE_RENDER_PASS)
			For Local light:ILightSceneNode = EachIn SceneLightList
				light.setVisible(False)
			Next
		EndIf
	EndMethod
	
	Method OnNodePreRender(node:ISceneNode) 
		CurrentSceneNode = node

		' This light manager only considers solid objects, but you are free To manipulate
		' lights during any phase, depending on your requirements.
		If (ESNRP_SOLID <> CurrentRenderPass) Then	Return

		' And in fact For this example, I only want To consider lighting For cube scene
		' nodes.  You will probably want To deal with lighting For (at least) mesh /
		' animated mesh scene nodes as well.
		If (node.getType() <> ESNT_CUBE) Then Return

		If (LIGHTS_NEAREST_NODE = Mode)
			' This is a naive implementation that prioritises every light in the scene
			' by its proximity To the node being rendered.  This produces some flickering
			' when lights orbit closer To a cube than its 'zone' lights.
			Local nodePosition:Vector3df = node.getAbsolutePosition()

			' Sort the light list by prioritising them based on their distance from the node
			' that's about to be rendered.
			Local sortingArray:TList = New TList

			For Local lightNode:ILightSceneNode = EachIn SceneLightList 
				Local distance:Double = lightNode.getAbsolutePosition().getDistanceFromSQ(nodePosition)				
				sortingArray.AddLast(LightDistanceElement.Create(lightNode, distance))
			Next

			sortingArray.sort(True, LightDistanceElement.SortFunc)

			' The list is now sorted by proximity To the node.
			' Turn on the three nearest lights, And turn the others off.
			Local cnt:Int = 0
			For Local element:LightDistanceElement = EachIn sortingArray
				element.node.setVisible(cnt < 3)
				cnt :+ 1
			Next

		ElseIf (LIGHTS_IN_ZONE = Mode)
			' Empty scene nodes are used To represent 'zones'.  For each solid mesh that
			' is being rendered, turn off all lights, Then find its 'zone' parent, and turn
			' on all lights that are found under that node in the scene graph.
			' This is a general purpose algorithm that doesn't use any special
			' knowledge of how this particular scene graph is organised.
			For Local lightNode:ILightSceneNode = EachIn SceneLightList 
				Local lightData:SLight = lightNode.getLightData()

				If (ELT_DIRECTIONAL <> lightData.getType()) Then lightNode.setVisible(False)
			Next

			Local parentZone:ISceneNode = findZone(node)
			If(parentZone) Then	turnOnZoneLights(parentZone)
		EndIf
	EndMethod

	Method OnNodePostRender(node:ISceneNode) 
		' I don't need to do any light management after individual node rendering.
	EndMethod


	' Find the empty scene node that is the parent of the specified node
	Method findZone:ISceneNode(node:ISceneNode)
		If Not node Then 	Return Null

		If (node.getType() = ESNT_EMPTY) Then Return node

		Return findZone(node.GetParent())
	EndMethod

	' Turn on all lights that are children (directly Or indirectly) of the
	' specified scene node.
	Method turnOnZoneLights(node:ISceneNode)
		Local children:ISceneNode[] = node.getChildren()
		
		For Local child:ISceneNode = EachIn children
			If (child.getType() = ESNT_LIGHT)
				child.setVisible(True)
			Else ' Assume that lights don't have any children that are also lights
				turnOnZoneLights(child)				
			EndIf
		Next
	EndMethod

	Method Init:CMyLightManager(smgr:ISceneManager)
		SceneManager = smgr
		Return Self
	EndMethod

	Function generate:ILightManager()
		Return New CMyLightManager
	EndFunction

	Method OnEvent:Int(event:SEvent)
		Local handled:Int = False

		If (event.getEventType()=EET_KEY_INPUT_EVENT And event.getKeyPressedDown())
			
			Local key:Int=event.getKeyInputKey()			
			handled = True
			
			Select key
				Case EKEY_KEY_1
					RequestedMode = NO_MANAGEMENT
				Case EKEY_KEY_2
					RequestedMode = LIGHTS_NEAREST_NODE
				Case EKEY_KEY_3
					RequestedMode = LIGHTS_IN_ZONE
				Default
					handled = False
			EndSelect
			
			If (NO_MANAGEMENT = RequestedMode)
				SceneManager.setLightManager(Null) ' Show that it's safe to register the light manager
			Else
				SceneManager.setLightManager(Self)
			EndIf
		EndIf

		Return handled
	EndMethod
		
	Method GetIEventReceiver:IEventReceiver()
		Return CMyLightManager_EventReceiver_Adapter(CMyLightManager_EventReceiver_Adapter.Create(CMyLightManager_EventReceiver_Adapter.generate)).Init(Self)
	EndMethod
EndType

' Because BlitzMax does not have multiple inheritance, we will use the adapter pattern to 
' implement the interface
Type CMyLightManager_EventReceiver_Adapter Extends IEventReceiver
	
	Field LightManager:CMyLightManager
	
	Method Init:CMyLightManager_EventReceiver_Adapter(LM:CMyLightManager)
		LightManager = LM				
		Return Self
	EndMethod

	Method OnEvent:Int(event:SEvent)
		Return LightManager.OnEvent(event)
	EndMethod

	' we must override the generate function in order for instantiation to work properly
	Function generate:IEventReceiver()
		Return New CMyLightManager_EventReceiver_Adapter
	EndFunction
EndType

' A utility class To aid in sorting scene nodes into a distance order
Type LightDistanceElement

	Field node:ILightSceneNode
	Field distance:Double

	Function Create:LightDistanceElement(n:ILightSceneNode, d:Double)
		Local element:LightDistanceElement = New LightDistanceElement
		element.node = n
		element.distance = d
		Return element
	EndFunction

	' Lower distance elements are sorted To the start of the array
	Function SortFunc:Int(element:Object, other:Object)
		Return (LightDistanceElement(element).distance > LightDistanceElement(other).distance)
	EndFunction	
EndType

Local driverChoice:String

If (AppArgs.length > 1)
	driverChoice = AppArgs[1]
Else
	driverChoice = Input("Please select the driver you want for this example:"+Chr(13)+Chr(10)+ ..
		" (a) Direct3D 9.0c"+Chr(13)+Chr(10)+" (b) Direct3D 8.1"+Chr(13)+Chr(10)+ ..
		" (c) OpenGL 1.5"+Chr(13)+Chr(10)+" (d) Software Renderer"+Chr(13)+Chr(10)+ ..
		" (e) Burning's Software Renderer"+Chr(13)+Chr(10)+" (f) NullDevice"+Chr(13)+Chr(10)+ ..
		" (otherKey) Exit"+Chr(13)+Chr(10)+Chr(13)+Chr(10))
EndIf

Local driverType:Int = EDT_NULL

Select driverChoice 
	Case "a" driverType=EDT_DIRECT3D9
	Case "b" driverType=EDT_DIRECT3D8
	Case "c" driverType=EDT_OPENGL
	Case "d" driverType=EDT_SOFTWARE
	Case "e" driverType=EDT_BURNINGVIDEO
	Case "f" driverType=EDT_NULL
	Default Return 
EndSelect

' create device And Exit If creation failed

Local device:IrrlichtDevice = ..
	IrrlichtDevice.Create(driverType, _DIMENSION2DI(640, 480), 32, False, False, False, Null)

If Not device Return ' could Not create selected driver.

Const lightRadius:Float = 60.0 ' Enough To reach the far side of each 'zone'

Local driver:IVideoDriver=device.getVideoDriver()
Local smgr:ISceneManager=device.getSceneManager()
Local guienv:IGUIEnvironment=device.getGUIEnvironment()

Local skin:IGUISkin = guienv.getSkin()
If (skin)
	skin.SetColor(EGDC_BUTTON_TEXT, _SCOLOR(255, 255, 255, 255))
	Local font:IGUIFont = guienv.getFont("../media/fontlucida.png")
	If (font) Then skin.setFont(font)
EndIf

guienv.addStaticText("1 - No light management", _RECTI(10,10,200,30))
guienv.addStaticText("2 - Closest 3 lights", _RECTI(10,30,200,50))
guienv.addStaticText("3 - Lights in zone", _RECTI(10,50,200,70))

' Add several "zones".  You could use this technique To light individual rooms, For example.
For Local zoneX:Float = -100 To 100 Step 50 
	For Local zoneY:Float = -60 To 60 Step 60
		' Start with an empty scene node, which we will use To represent a zone.
		Local zoneRoot:ISceneNode = smgr.addEmptySceneNode()
		zoneRoot.setPosition(_VECTOR3DF(zoneX, zoneY, 0))

		' Each zone contains a rotating cube
		Local node:IMeshSceneNode = smgr.addCubeSceneNode(15, zoneRoot)
		Local rotation:ISceneNodeAnimator = smgr.createRotationAnimator(_VECTOR3DF(0.25, 0.5, 0.75))
		node.addAnimator(rotation)
		rotation.drop()

		' And each cube has three lights attached To it.  The lights are attached To billboards so
		' that we can see where they are.  The billboards are attached To the cube, so that the
		' lights are indirect descendents of the same empty scene node as the cube.
		Local billboard:IBillboardSceneNode = smgr.addBillboardSceneNode(node)
		billboard.setPosition(_VECTOR3DF(0, -14, 30))
		billboard.setMaterialType(EMT_TRANSPARENT_ADD_COLOR )
		billboard.setMaterialTexture(0, driver.getTexture("../media/particle.bmp"))
		billboard.setMaterialFlag(EMF_LIGHTING, False)
		Local light:ILightSceneNode = smgr.addLightSceneNode(billboard, _VECTOR3DF(0, 0, 0), _SCOLORF(1, 0, 0), lightRadius)

		billboard = smgr.addBillboardSceneNode(node)
		billboard.setPosition(_VECTOR3DF(-21, -14, -21))
		billboard.setMaterialType(EMT_TRANSPARENT_ADD_COLOR )
		billboard.setMaterialTexture(0, driver.getTexture("../media/particle.bmp"))
		billboard.setMaterialFlag(EMF_LIGHTING, False)
		light = smgr.addLightSceneNode(billboard, _VECTOR3DF(0, 0, 0), _SCOLORF(0, 1, 0), lightRadius)

		billboard = smgr.addBillboardSceneNode(node)
		billboard.setPosition(_VECTOR3DF(21, -14, -21))
		billboard.setMaterialType(EMT_TRANSPARENT_ADD_COLOR )
		billboard.setMaterialTexture(0, driver.getTexture("../media/particle.bmp"))
		billboard.setMaterialFlag(EMF_LIGHTING, False)
		light = smgr.addLightSceneNode(billboard, _VECTOR3DF(0, 0, 0), _SCOLORF(0, 0, 1), lightRadius)

		' Each cube also has a smaller cube rotating around it, To show that the cubes are being
		' lit by the lights in their 'zone', not just lights that are their direct children.
		node = smgr.addCubeSceneNode(5, node)
		node.setPosition(_VECTOR3DF(0, 21, 0))
	Next
Next

smgr.addCameraSceneNode(Null, _VECTOR3DF(0,0,-130), _VECTOR3DF(0,0,0))

Local myLightManager:ILightManager = CMyLightManager(CMyLightManager.Create(CMyLightManager.generate)).Init(smgr)

smgr.setLightManager(Null) ' This is the Default: we won't do light management until told to do it.
device.setEventReceiver(CMyLightManager(myLightManager).GetIEventReceiver())

Local lastFPS:Int = -1

While(device.run())
	If (device.isWindowActive())
		driver.beginScene(True, True, _SCOLOR(255,100,101,140))
	
		smgr.drawAll()
		guienv.drawAll()
		
		driver.endScene()
	
		Local fps:Int = driver.getFPS()
	
		If (lastFPS <> fps)
		  Local str:String = "Managed Lights ["
		  str :+ driver.getName()
		  str :+ "] FPS:"
		  str :+ fps
	
		  device.setWindowCaption(str)
		  lastFPS = fps
		EndIf
	EndIf
EndWhile

myLightManager.drop() ' Drop my implicit reference
device.drop()
